﻿using System;
using System.Collections.Generic;
using System.Diagnostics.Contracts;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace GV
{
    public class Article
    {
        public string Title { get; set; }
        public int Start { get; set; }
        public int Length { get; set; }
        public byte[] Data { get; set; }

        public override string ToString()
        {
            return Title;
        }
    }

    internal class Program
    {
        static void Main(string[] args)
        {
            Console.WriteLine("Grapevine article ripper - Issues 14 and above");
            Console.WriteLine("");

            if (args.Length == 0)
            {
                Console.WriteLine("Usage: <article file>");
                Console.WriteLine("");
                Console.WriteLine("Articles will be extracted in the current folder");
                Console.WriteLine("");
                return;
            }

            Console.WriteLine($"Extracting {args[0]}");
            Console.WriteLine("");


            var data = File.ReadAllBytes(args[0]);
            using (var reader = new BigEndianReader(new MemoryStream(data)))
            {
                var articleCount = reader.ReadUInt16();

                Console.WriteLine($"Article count: {articleCount}");
                Console.WriteLine("");

                var articles = new List<Article>();

                for (var i = 0; i < articleCount; i++)
                {
                    var article = new Article
                    {
                        Title = reader.ReadAscii(0x28),
                        Start = reader.ReadInt32(),
                        Length = reader.ReadInt32()
                    };

                    articles.Add(article);
                }

                foreach (var article in articles)
                {
                    var tempdata = data.Skip(article.Start).Take(article.Length).ToArray();
                    using (var artreader = new BigEndianReader(new MemoryStream(tempdata)))
                    {
                        var header = artreader.ReadAscii(8);
                        if (header == "GARTTEXT")
                        {
                            var artsize = artreader.ReadInt32();
                            article.Data = artreader.ReadBytes(artsize);
                            article.Data[0] = (byte)'P';
                            article.Data[1] = (byte)'P';
                            article.Data[2] = (byte)'2';
                            article.Data[3] = (byte)'0';
                            var filename = ReplaceInvalidChars(article.Title.Trim()).Replace(' ', '_') + ".pp20";
                            Console.WriteLine($"Extracting article - {filename}");
                            File.WriteAllBytes(filename, article.Data);
                        }
                    }

                }
            }

            Console.WriteLine("!!DONE!!");
        }

        public static string ReplaceInvalidChars(string filename)
        {
            return string.Join("_", filename.Split(Path.GetInvalidFileNameChars()));
        }
    }

    public class BigEndianReader : BinaryReader
    {
        public BigEndianReader(Stream stream) : base(stream) { }

        public override int ReadInt32()
        {
            var data = base.ReadBytes(4);
            Array.Reverse(data);
            return BitConverter.ToInt32(data, 0);
        }

        public override Int16 ReadInt16()
        {
            var data = base.ReadBytes(2);
            Array.Reverse(data);
            return BitConverter.ToInt16(data, 0);
        }

        public override UInt16 ReadUInt16()
        {
            var data = base.ReadBytes(2);
            Array.Reverse(data);
            return BitConverter.ToUInt16(data, 0);
        }

        public override Int64 ReadInt64()
        {
            var data = base.ReadBytes(8);
            Array.Reverse(data);
            return BitConverter.ToInt64(data, 0);
        }

        public override UInt32 ReadUInt32()
        {
            var data = base.ReadBytes(4);
            Array.Reverse(data);
            return BitConverter.ToUInt32(data, 0);
        }

        public string ReadAscii(int length)
        {
            var data = base.ReadBytes(length);
            return Encoding.ASCII.GetString(data);
        }

        public bool EndOfStream()
        {
            if (BaseStream.Position == BaseStream.Length)
            {
                return true;
            }
            return false;
        }
    }
}
